# w.c.s. - web application for online forms
# Copyright (C) 2005-2022  Entr'ouvert
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import json

from django.template.loader import render_to_string
from django.utils.timezone import now
from quixote import get_publisher, get_request, get_response, get_session, redirect
from quixote.directory import Directory
from quixote.html import TemplateIO, htmltext

from wcs.backoffice.management import FormBackofficeEditPage, FormBackOfficeStatusPage
from wcs.forms.common import FormStatusPage
from wcs.qommon import _, misc, template
from wcs.qommon.afterjobs import AfterJob
from wcs.qommon.backoffice.listing import pagination_links
from wcs.qommon.backoffice.menu import html_top
from wcs.qommon.errors import TraversalError
from wcs.qommon.form import FileWidget, Form, SingleSelectWidget, StringWidget
from wcs.qommon.storage import Equal, StrictNotEqual
from wcs.testdef import TestDef, TestError, TestResult


class TestPage(FormBackOfficeStatusPage):
    _q_extra_exports = ['delete', 'export', 'edit', ('edit-data', 'edit_data')]

    def __init__(self, component, objectdef):
        try:
            self.testdef = TestDef.get(component)
        except KeyError:
            raise TraversalError()

        filled = self.testdef.build_formdata(objectdef, include_fields=True)
        super().__init__(objectdef, filled)

    def _q_index(self):
        get_response().add_javascript(['select2.js'])
        return super()._q_index()

    def _q_traverse(self, path):
        get_response().breadcrumb.append((str(self.testdef.id) + '/', str(self.testdef)))
        return super(FormStatusPage, self)._q_traverse(path)

    def html_top(self, title=None):
        html_top('tests', '%s - %s' % (self.formdef.name, _('Tests')))

    def should_fold_summary(self, mine, request_user):
        return False

    def get_extra_context_bar(self, parent=None):
        return render_to_string('wcs/backoffice/test_sidebar.html', context={})

    def status(self):
        r = TemplateIO(html=True)
        r += htmltext('<h2>%s</h2>') % self.testdef
        r += self.receipt(always_include_user=True, mine=False)
        return r.getvalue()

    def delete(self):
        form = Form(enctype='multipart/form-data')
        form.add_submit('delete', _('Submit'))
        form.add_submit('cancel', _('Cancel'))

        if form.get_widget('cancel').parse():
            return redirect('.')
        if not form.is_submitted() or form.has_errors():
            get_response().breadcrumb.append(('delete', _('Delete')))
            r = TemplateIO(html=True)
            r += htmltext('<h2>%s %s</h2>') % (_('Deleting Test:'), self.testdef)
            r += form.render()
            return r.getvalue()
        else:
            TestDef.remove_object(self.testdef.id)
            return redirect('..')

    def export(self):
        get_response().set_content_type('application/json')
        get_response().set_header(
            'content-disposition', 'attachment; filename=wcs_test_%s.json' % self.testdef.name
        )
        return json.dumps(self.testdef.export_to_json())

    def edit_data(self):
        self.filled.data['edited_testdef_id'] = self.testdef.id

        f = FormBackofficeEditPage(self.formdef.url_name)
        f.testdef = self.testdef
        f.edited_data = self.filled
        f.action_url = 'edit-data'

        return f._q_index()

    def edit(self):
        form = Form(enctype='multipart/form-data')
        form.add(StringWidget, 'name', title=_('Name'), required=True, size=50, value=self.testdef.name)

        user_options = [('', '---', '')] + [
            (x.id, str(x), x.id) for x in get_publisher().user_class.select(order_by='name')
        ]
        form.add(
            SingleSelectWidget,
            'user',
            title=_('User'),
            value=self.testdef.data['user'].get('id', ''),
            options=user_options,
            **{'data-autocomplete': 'true'},
        )

        form.add_submit('submit', _('Submit'))
        form.add_submit('cancel', _('Cancel'))

        if form.get_widget('cancel').parse():
            return redirect('.')
        if not form.is_submitted() or form.has_errors():
            get_response().breadcrumb.append(('edit', _('Edit test')))
            r = TemplateIO(html=True)
            r += htmltext('<h2>%s</h2>') % (_('Edit test'))
            r += form.render()
            return r.getvalue()
        else:
            self.testdef.name = form.get_widget('name').parse()

            user_id = form.get_widget('user').parse()
            if user_id:
                user = get_publisher().user_class.get(user_id)
                self.testdef.data['user'] = user.get_json_export_dict()
            else:
                self.testdef.data['user'] = None

            self.testdef.store()
            return redirect('.')


class TestsDirectory(Directory):
    _q_exports = ['', 'new', ('import', 'p_import'), 'results']
    section = 'tests'

    def __init__(self, objectdef):
        self.objectdef = objectdef
        self.results = TestResultsDirectory(objectdef)

    def _q_traverse(self, path):
        get_response().breadcrumb.append(('tests/', _('Tests')))
        html_top('tests', '%s - %s' % (self.objectdef.name, _('Tests')))
        return super()._q_traverse(path)

    def _q_lookup(self, component):
        return TestPage(component, self.objectdef)

    def _q_index(self):
        context = {
            'testdefs': TestDef.select(
                [Equal('object_type', self.objectdef.get_table_name()), Equal('object_id', self.objectdef.id)]
            ),
            'formdata': self.objectdef.data_class().select([StrictNotEqual('status', 'draft')]),
            'has_deprecated_fields': any(
                x.type in ('table', 'table-select', 'tablerows', 'ranked-items')
                for x in self.objectdef.fields
            ),
        }
        get_response().add_javascript(['popup.js'])
        return template.QommonTemplateResponse(
            templates=['wcs/backoffice/tests.html'], context=context, is_django_native=True
        )

    def new(self):
        form = Form(enctype='multipart/form-data')
        form.add(StringWidget, 'name', title=_('Name'), required=True, size=50)
        formadata_options = [
            (x.id, '%s - %s' % (x.id_display, x.user or _('Unknown User')))
            for x in self.objectdef.data_class().select(
                [StrictNotEqual('status', 'draft')], order_by='-receipt_time'
            )
        ]
        form.add(SingleSelectWidget, 'formdata_id', title=_('Form'), required=True, options=formadata_options)
        form.add_submit('submit', _('Submit'))
        form.add_submit('cancel', _('Cancel'))

        if form.get_widget('cancel').parse():
            return redirect('.')

        if not form.is_submitted() or form.has_errors():
            get_response().breadcrumb.append(('new', _('New')))
            r = TemplateIO(html=True)
            r += htmltext('<h2>%s</h2>') % _('New test')
            r += form.render()
            return r.getvalue()
        else:
            formdata_id = form.get_widget('formdata_id').parse()
            formdata = self.objectdef.data_class().get(formdata_id)

            testdef = TestDef.create_from_formdata(self.objectdef, formdata)
            testdef.name = form.get_widget('name').parse()
            testdef.store()

        return redirect('.')

    def p_import(self):
        form = Form(enctype='multipart/form-data')

        form.add(FileWidget, 'file', title=_('File'), required=True)
        form.add_submit('submit', _('Import Test'))
        form.add_submit('cancel', _('Cancel'))

        if form.get_submit() == 'cancel':
            return redirect('.')

        if form.is_submitted() and not form.has_errors():
            try:
                return self.import_submit(form)
            except ValueError:
                pass

        get_response().breadcrumb.append(('import', _('Import')))
        r = TemplateIO(html=True)
        r += htmltext('<h2>%s</h2>') % _('Import Test')
        r += htmltext('<p>%s</p>') % _(
            'You can add a new test or update an existing one by importing a JSON file.'
        )
        r += form.render()
        return r.getvalue()

    def import_submit(self, form):
        fp = form.get_widget('file').parse().fp

        try:
            testdef = TestDef.import_from_json(json.loads(fp.read()))
        except Exception as e:
            form.set_error('file', str(e))
            raise ValueError()

        get_session().message = ('info', _('Test "%s" has been successfully imported.') % testdef.name)
        return redirect('.')


class TestResultPage(Directory):
    _q_exports = ['']

    def __init__(self, component, objectdef):
        try:
            self.test_result = TestResult.get(component)
        except KeyError:
            raise TraversalError()

        self.objectdef = objectdef

    def _q_traverse(self, path):
        get_response().breadcrumb.append(
            (str(self.test_result.id) + '/', _('Result #%s') % self.test_result.id)
        )
        return super()._q_traverse(path)

    def _q_index(self):
        testdefs = TestDef.select(
            [Equal('object_type', self.objectdef.get_table_name()), Equal('object_id', self.objectdef.id)]
        )
        testdefs_by_id = {x.id: x for x in testdefs}
        for test in self.test_result.results:
            if test['id'] in testdefs_by_id:
                test['url'] = testdefs_by_id[test['id']].get_admin_url()

        return template.QommonTemplateResponse(
            templates=['wcs/backoffice/test-result.html'],
            context={'test_result': self.test_result},
            is_django_native=True,
        )


class TestResultsDirectory(Directory):
    _q_exports = ['', 'run']
    section = 'test_results'

    def __init__(self, objectdef):
        self.objectdef = objectdef

    def _q_traverse(self, path):
        get_response().breadcrumb.append(('results/', _('Test results')))
        html_top('test-results', '%s - %s' % (self.objectdef.name, _('Test results')))
        return super()._q_traverse(path)

    def _q_lookup(self, component):
        return TestResultPage(component, self.objectdef)

    def _q_index(self):
        criterias = [
            Equal('object_type', self.objectdef.get_table_name()),
            Equal('object_id', self.objectdef.id),
        ]

        offset = misc.get_int_or_400(get_request().form.get('offset', 0))
        limit = misc.get_int_or_400(get_request().form.get('limit', 25))
        total_count = TestResult.count(criterias)

        context = {
            'test_results': TestResult.select(criterias, offset=offset, limit=limit, order_by='-id'),
            'has_testdefs': bool(TestDef.count(criterias)),
            'pagination_links': pagination_links(offset, limit, total_count, load_js=False),
        }
        return template.QommonTemplateResponse(
            templates=['wcs/backoffice/test-results.html'], context=context, is_django_native=True
        )

    def run(self):
        test_result = TestsAfterJob.run_tests(self.objectdef, _('Manual run.'))
        return redirect(test_result.get_admin_url())


class TestsAfterJob(AfterJob):
    def __init__(self, objectdef, reason, snapshot=None, **kwargs):
        super().__init__(
            objectdef_class=objectdef.__class__,
            objectdef_id=objectdef.id,
            reason=reason,
            snapshot_id=snapshot.id if snapshot else None,
            **kwargs,
        )

    def execute(self):
        objectdef = self.kwargs['objectdef_class'].get(self.kwargs['objectdef_id'])
        reason = self.kwargs['reason']

        result = self.run_tests(objectdef, reason)

        if result and self.kwargs['snapshot_id'] is not None:
            snapshot = get_publisher().snapshot_class.get(self.kwargs['snapshot_id'])
            snapshot.test_result_id = result.id
            snapshot.store()

    @staticmethod
    def run_tests(objectdef, reason):
        testdefs = TestDef.select(
            [Equal('object_type', objectdef.get_table_name()), Equal('object_id', objectdef.id)]
        )
        if not testdefs:
            return

        for test in testdefs:
            try:
                test.run(objectdef)
            except TestError as e:
                test.error = str(e)

        test_result = TestResult()
        test_result.object_type = objectdef.get_table_name()
        test_result.object_id = objectdef.id
        test_result.timestamp = now()
        test_result.success = not any(hasattr(test, 'error') for test in testdefs)
        test_result.reason = str(reason)
        test_result.results = [
            {
                'id': test.id,
                'name': str(test),
                'error': getattr(test, 'error', None),
            }
            for test in testdefs
        ]
        test_result.results.sort(key=lambda x: bool(x['error']))
        test_result.store()

        return test_result
